from typing import Any

import numpy as np
from dataclasses import dataclass

from ..data import BeamData
from ..core import Processor
from ..utils import as_scipy_csr, as_scipy_coo, as_numpy, as_tensor


@dataclass
class Similarities:
    index: Any
    distance: Any
    values: Any = None
    sparse_scores: Any = None
    metric: str = None
    model: str = None


class BeamSimilarity(Processor):

    def __init__(self, *args, metric=None, **kwargs):
        super().__init__(*args, metric=metric, **kwargs)
        self.metric = self.get_hparam('metric', metric)
        self.index = None
        self._is_trained = None
        self.reset()

    @property
    def is_trained(self):
        return self._is_trained

    def reset(self):
        self.index = None
        self._is_trained = False

    @staticmethod
    def extract_data_and_index(x, index=None, convert_to='numpy'):
        if isinstance(x, BeamData) or hasattr(x, 'beam_class') and x.beam_class == 'BeamData':
            index = x.index
            x = x.values

        if convert_to == 'numpy':
            x = as_numpy(x)
        elif convert_to == 'tensor':
            x = as_tensor(x)
        elif convert_to == 'scipy_csr':
            x = as_scipy_csr(x)
        elif convert_to == 'scipy_coo':
            x = as_scipy_coo(x)
        else:
            raise ValueError(f"Unknown conversion: {convert_to}")

        return x, as_numpy(index)

    @property
    def metric_type(self):
        return self.metric

    def add(self, x, index=None, **kwargs):
        raise NotImplementedError

    def search(self, x, k=1):
        raise NotImplementedError

    def train(self, x):
        raise NotImplementedError

    def remove_ids(self, ids):
        raise NotImplementedError

    def reconstruct(self, id0):
        raise NotImplementedError

    def reconstruct_n(self, id0, id1):
        raise NotImplementedError

    @property
    def ntotal(self):
        if self.index is not None:
            return len(self.index)
        return 0

    def __len__(self):
        return self.ntotal

    def save_state(self, path, ext=None, **kwargs):
        state = {attr: getattr(self, attr) for attr in self.exclude_pickle_attributes}
        state['hparams'] = self.hparams
        bd = BeamData(state, path=path)
        bd.store(**kwargs)

    def load_state(self, path, ext=None, **kwargs):
        bd = BeamData(path=path)
        state = bd.cache(**kwargs).values
        for attr in self.exclude_pickle_attributes:
            setattr(self, attr, state[attr])

    def get_index(self, index):
        return self.index[as_numpy(index)]

    def add_index(self, x, index=None):

        if self.index is None:
            if index is None:
                index = np.arange(len(x))
            else:
                index = as_numpy(index)
            self.index = index
        else:
            if index is None:
                index = np.arange(len(x)) + self.index.max() + 1
            else:
                index = as_numpy(index)
            self.index = np.concatenate([self.index, index])




from .resource import beam_llm
from .utils import *

from .tools import LLMTool, LLMToolProperty
from .task import LLMTask

from .simulators import openai as openai_simulator
from .simulators import text_generation as tgi_simulator
from .simulators import openai_legacy as openai_legacy_simulator




from ..config import BeamConfig, TransformerConfig, BeamParam


class SimilarityConfig(BeamConfig):

    parameters = [
        BeamParam('vector_dimension', int, None, 'dimension of the vectors'),
        BeamParam('expected_population', int, int(1e6), 'expected population of the index'),
        BeamParam('metric', str, 'l2', 'distance metric [l2, cosine, max, prod, quantile]'),
        BeamParam('training_device', str, 'cpu', 'device for training'),
        BeamParam('inference_device', str, 'cpu', 'device for inference'),
        BeamParam('ram_footprint', int, 2 ** 8 * int(1e9), 'RAM footprint'),
        BeamParam('gpu_footprint', int, 24 * int(1e9), 'GPU footprint'),
        BeamParam('exact', bool, False, 'exact search'),
        BeamParam('nlists', int, None, 'number of lists for IVF'),
        BeamParam('faiss_M', int, None, 'M for IVFPQ'),
        BeamParam('reducer', str, 'umap', 'dimensionality reduction method'),
        BeamParam('quantile', float, 0.9, 'quantile for the quantile metric'),
    ]


class TFIDFConfig(TransformerConfig):

    defaults = dict(metric='bm25')

    parameters = [
        BeamParam('max_features', int, 1000, 'maximum number of features'),
        BeamParam('max_df', float, 0.95, 'maximum document frequency'),
        BeamParam('min_df', float, 2, 'minimum document frequency'),
        BeamParam('use_idf', bool, True, 'use inverse document frequency'),
        BeamParam('smooth_idf', bool, True, 'smooth inverse document frequency'),
        BeamParam('sublinear_tf', bool, False, 'apply sublinear term frequency scaling'),
        BeamParam('sparse_framework', str, 'torch', 'sparse framework, can be "torch" or "scipy"'),
        BeamParam('sparse_layout', str, 'coo', 'sparse layout, can be "coo" or "csr"'),
        BeamParam('norm', str, 'l2', 'Each output row will have unit norm, either [l1, l2, none]'),
        BeamParam('bm25_k1', float, 1.5, 'bm25 k1 parameter'),
        BeamParam('bm25_b', float, 0.75, 'bm25 b parameter'),
        BeamParam('bm25_epsilon', float, 0.25, 'bm25 epsilon parameter'),
    ]




import time
import numpy as np
import os

import torch
from collections import defaultdict
import pandas as pd
from contextlib import contextmanager
from timeit import default_timer as timer
import threading

from ..utils import (pretty_format_number, as_numpy, pretty_print_timedelta, recursive_flatten, rate_string_format,
                     nested_defaultdict, as_tensor, squeeze_scalar, check_type, check_element_type, lazy_property,
                     strip_prefix, recursive_detach, recursive_to_cpu)

from ..utils import tqdm_beam as tqdm
from ..logger import beam_logger as logger
from ..data import BeamData


class BeamReport(object):

    def __init__(self, objective=None, objective_mode='max', aux_objectives=None, aux_objectives_modes=None):

        self.scalar = None
        self.aux = None

        self.scalars = None
        self.buffer = None

        self.scalar_kwargs = None
        self.scalars_kwargs = None
        self.aux_kwargs = None

        self.scalar_aggregation = None
        self.scalars_aggregation = None

        if aux_objectives is None:
            aux_objectives = []
        if aux_objectives_modes is None:
            aux_objectives_modes = []

        if objective is not None:
            aux_objectives.insert(0, objective)
            aux_objectives_modes.insert(0, objective_mode)

        self.objective_names = aux_objectives
        self.objectives_modes = aux_objectives_modes

        self.objective_name = None
        self.objective_mode = None

        self.epoch = None
        self.best_epoch = None
        self.objective = None
        self.best_objective = None
        self.subset_context = None
        self.best_state = False
        self.iteration = None
        self._data = None
        self.state = None
        self.t0 = time.time()
        self.n_epochs = None
        self.total_time = None
        self.estimated_time = None
        self.first_epoch = None
        self.subsets_keys = None
        self.batch_size_context = None
        self.total_epochs = None
        self.stack = []
        self.stack_size = 4

    def info(self, msg, new=False):
        logger.info(msg)
        if new:
            self.stack.append(msg)
        else:
            self.stack[-1] = f"{self.stack[-1]}\n{msg}"
        self.stack = self.stack[-self.stack_size:]

    def llm_info(self):
        logs = "\n\n".join(self.stack)
        prompt = f"You are an ML expert. You need to interpret and analyze logs of deep learning training runs. \n" \
                 f"The logs contains metrics and reports of the training process. \n" \
                 f"Please provide an analysis and suggest solutions for any problems you find. For example: \n" \
                 f"overfitting, underfitting, etc. Be concise and respond in less than 100 words \n" \
                 f"========================================================================\n\n" \
                 f"These are the experiment logs: \n\n" \
                 f"{logs}\n\n" \
                 f"Response: \"\"\"\n{{text input here}}\n\"\"\""

        llm_response = self.llm.ask(prompt)
        print()
        logger.info(f"LLM response: {llm_response.text}")

    def reset_epoch(self, epoch, total_epochs=None):

        self.state = 'before_epoch'
        self.scalar = defaultdict(list)
        self.aux = defaultdict(dict)

        self.scalars = nested_defaultdict(list)
        self.buffer = nested_defaultdict(list)

        self.scalar_kwargs = {}
        self.scalars_kwargs = {}
        self.aux_kwargs = defaultdict(dict)

        self.scalar_aggregation = {}
        self.scalars_aggregation = {}

        self.subsets_keys = nested_defaultdict(list)
        self.epoch = epoch
        if total_epochs is None:
            total_epochs = epoch + 1
        self.total_epochs = total_epochs

    def reset_time(self, first_epoch=None, n_epochs=None):
        self.t0 = time.time()
        self.n_epochs = n_epochs
        if first_epoch is None:
            first_epoch = 0
        self.first_epoch = first_epoch
        self.estimated_time = None
        self.total_time = None

    @property
    def data(self):
        if self._data is None:
            data = nested_defaultdict(dict)
            for k, v in self.scalar.items():
                subset, name = self.extract_subset_and_name(k)
                if subset is None:
                    data['scalar'][name] = v
                else:
                    data[subset]['scalar'][name] = v
            for k, v in self.scalars.items():
                subset, name = self.extract_subset_and_name(k)
                if subset is None:
                    data['scalars'][name] = v
                else:
                    data[subset]['scalars'][name] = v
            for dtype, v_dtype  in self.aux.items():
                for k, v in v_dtype.items():
                    subset, name = self.extract_subset_and_name(k)
                    if subset is None:
                        data[dtype][name] = v
                    else:
                        data[subset][dtype][name] = v

            data['global'] = {'epoch': self.epoch,
                              'best_epoch': self.best_epoch,
                              'objective': self.objective,
                              'best_objective': self.best_objective,
                              'total_time': self.total_time,
                              'estimated_time': self.estimated_time,
                              }

            data['objective'] = self.objective

            self._data = BeamData(data)

        return self._data

    def write_to_path(self, path):
        self.data.store(path=path)

    @staticmethod
    def format_stat(k, v):
        format = f"{k if k != 'mean' else 'avg'}:{pretty_format_number(v)}".ljust(15)
        return format

    @staticmethod
    def format(v):
        v_type = check_element_type(v)
        if v_type == 'int':
            if v >= 1000:
                return f"{float(v): .4}"
            else:
                return str(v)
        elif v_type == 'float':
            return f"{v: .4}"
        else:
            return v

    @lazy_property
    def llm(self):
        from ..config import get_beam_llm
        return get_beam_llm()

    def print_stats(self):

        for subset, data_keys in self.subsets_keys.items():

            self.info(f'{subset}:')

            if 'stats' in data_keys:
                stats = data_keys['stats']
                self.info('| '.join([f"{k}: {BeamReport.format(self.aux['stats'][f'{subset}/{k}'])} " for k in stats]))

            if 'scalar' in data_keys:
                scalars = data_keys['scalar']
                for k in scalars:
                    v = self.scalar[f'{subset}/{k}']

                    v_type = check_type(v)
                    if v_type.major != 'scalar' and np.var(as_numpy(v)) > 0:
                        stat = pd.Series(v, dtype=np.float32).describe()
                    else:
                        if v_type.major != 'scalar':
                            v = v[0]
                        v = int(v) if v_type.element == 'int' else float(v)
                        stat = {'val': v}

                    stat = '| '.join([BeamReport.format_stat(k, v) for k, v in dict(stat).items() if k != 'count'])

                    if len(k) > 11:
                        paramp = f'{k[:4]}...{k[-4:]}:'
                    else:
                        paramp = f'{k}:'

                    self.info(f'{paramp: <12} | {stat}')

        if self.llm is not None:
            threading.Thread(target=self.llm_info).start()

    def print_metadata(self):

        self.info('----------------------------------------------------------'
                    '---------------------------------------------------------------------', new=not self.best_state)
        objective_str = ''
        if self.best_objective is not None:
            objective_str = f"Current objective: {pretty_format_number(self.objective)} " \
                            f"(Best objective: {pretty_format_number(self.best_objective)} " \
                            f" at epoch {self.best_epoch + 1})"

        if self.epoch is not None:
            done_epochs = self.epoch - self.first_epoch
            self.info(f'Finished epoch {done_epochs + 1}/{self.n_epochs} (Total trained epochs {self.total_epochs}). '
                        f'{objective_str}')

        if self.total_time is not None:
            total_time = pretty_print_timedelta(self.total_time)
            estimated_time = pretty_print_timedelta(self.estimated_time)
            self.info(f'Elapsed time: {total_time}. Estimated remaining time: {estimated_time}.', )

    def write_to_tensorboard(self, writer, hparams=None):

        metrics = {}
        for k, v in self.scalar.items():
            kwargs = self.scalar_kwargs.get(k, {})
            agg = self.scalar_aggregation.get(k, 'mean')
            v = self.aggregate_scalar(v, agg)
            if agg != 'mean':
                k = f'{k}_{agg}'
            writer.add_scalar(k, v, **kwargs)
            metrics[k] = v

        if hparams is not None and len(metrics):
            writer.add_hparams(hparams, metrics, name=os.path.join('..', 'hparams'), global_step=self.epoch)

        for k, v in self.scalars.items():

            v_agg = {}
            kwargs = self.scalars_kwargs.get(k, {})
            agg = self.scalars_aggregation.get(k, 'mean')

            for kk, vv in v.items():
                v_agg[kk] = self.aggregate_scalar(vv, agg)

            writer.add_scalars(k, v_agg, **kwargs)

        for k, v in self.aux.items():
            if hasattr(writer, f'add_{k}'):
                writer_func = getattr(writer, f'add_{k}')
                for kk, vv in v.items():
                    kwargs = self.aux_kwargs.get(k, {}).get(kk, {})
                    writer_func(kk, vv, **kwargs)

    @staticmethod
    def extract_subset_and_name(k):
        if '/' not in k:
            subset = None
            name = k
        else:
            subset = k.split('/')[0]
            name = strip_prefix(k, f"{subset}/")
        return subset, name

    @lazy_property
    def comparison(self):
        return {'max': np.greater, 'min': np.less}[self.objective_mode]

    def set_objective(self, objective):

        self.objective = objective

        if self.best_objective is None or self.comparison(self.objective, self.best_objective):
            self.info(f"Epoch {self.epoch+1}: The new best objective is {pretty_format_number(objective)}", new=True)
            self.best_objective = objective
            self.best_epoch = self.epoch
            self.best_state = True
        else:
            self.best_state = False

    def set_objective_name(self, keys):

        for i, o in enumerate(self.objective_names):
            if o is not None and o in keys:
                self.objective_name = o
                self.objective_mode = self.objectives_modes[i]
                return

    @contextmanager
    def track_epoch(self, subset, batch_size=None, training=True):
        self.subset_context = subset
        self.batch_size_context = batch_size
        self.state = 'in_epoch'

        t0 = timer()
        yield
        delta = timer() - t0
        n_iter = self.iteration + 1
        track_objective = not training

        self.state = 'after_epoch'

        self.report_data('seconds', delta, data_type='stats')
        self.report_data('batches', n_iter, data_type='stats')
        self.report_data('samples', n_iter * batch_size, data_type='stats')
        self.report_data('batch_rate', rate_string_format(n_iter, delta), data_type='stats')
        self.report_data('sample_rate', rate_string_format(n_iter * batch_size, delta), data_type='stats')

        self.total_time = time.time() - self.t0
        if (self.n_epochs is not None) and (self.n_epochs is not None) and (self.n_epochs > 0):

            n_epochs = self.n_epochs - self.first_epoch
            epoch = self.epoch - self.first_epoch

            self.estimated_time = self.total_time * (n_epochs - epoch - 1) / (epoch + 1)

        agg = None

        if self.objective_name is None:
            self.set_objective_name(list(self.subsets_keys[subset]['scalar']))

        for name in self.subsets_keys[subset]['scalar']:

            k = f'{subset}/{name}' if subset is not None else name
            v = self.scalar[k]

            self.scalar[k] = self.stack_scalar(v, batch_size=batch_size)

            if name == self.objective_name and track_objective:
                agg = self.scalar_aggregation.get(k, None)
                self.set_objective(self.aggregate_scalar(self.scalar[k], agg))

        for data_type in self.subsets_keys[subset]:
            if data_type in ['scalar', 'scalars', 'stats']:
                continue
            for name in self.subsets_keys[subset][data_type]:
                k = f'{subset}/{name}' if subset is not None else name
                v = self.aux[data_type][k]
                self.aux[data_type][k] = recursive_to_cpu(v)

        if self.objective_name and track_objective and agg is None:
            logger.warning(f"The objective {self.objective_name} is missing from the validation results")

        for name in self.subsets_keys[subset]['scalars']:
            k = f'{subset}/{name}' if subset is not None else name
            v = self.scalar[k]
            for kk, vv in v.items():
                self.scalars[k][kk] = self.stack_scalar(vv, batch_size=batch_size)

        self.subset_context = None
        self.batch_size_context = None

    def iterate(self, generator, **kwargs):
        for i, batch in tqdm(generator, **kwargs):
            self.iteration = i
            yield i, batch

    @staticmethod
    def detach_scalar(val):
        recursive_detach(val)
        return val

    # @staticmethod
    # def detach_scalar(val):
    #     val_type = check_type(val)
    #     if val_type.major == 'scalar':
    #         if val_type.element == 'float':
    #             val = float(val)
    #             pass
    #         elif val_type.element == 'int':
    #             val = int(val)
    #     elif val_type.minor == 'tensor':
    #         val = val.detach().cpu()
    #     elif val_type.major == 'container':
    #         val = as_tensor(val, device='cpu')
    #
    #     return val

    @staticmethod
    def stack_scalar(val, batch_size=None):

        val_type = check_type(val)

        if val_type.major == 'container' and val_type.minor == 'list':

            v_minor = check_type(val[0]).minor
            if v_minor == 'tensor':
                oprs = {'cat': torch.cat, 'stack': torch.stack}
            elif v_minor == 'numpy':
                oprs = {'cat': np.concatenate, 'stack': np.stack}
            elif v_minor == 'pandas':
                oprs = {'cat': pd.concat, 'stack': pd.concat}
            elif v_minor == 'native':
                oprs = {'cat': torch.tensor, 'stack': torch.tensor}
            else:
                oprs = {'cat': lambda x: x, 'stack': lambda x: x}

            opr = oprs['cat']
            if batch_size is not None and hasattr(val[0], '__len__') \
                    and len(val[0]) != batch_size and len(val[0]) == len(val[-1]):
                opr = oprs['stack']

            val = opr(val)

        elif val_type.major == 'array' and val_type.minor == 'list' and val_type.element in ['int', 'float']:
            val = as_tensor(val, device='cpu')

        val = squeeze_scalar(val)

        return val

    @staticmethod
    def aggregate_scalar(val, aggregation, batch_size=None):

        val = BeamReport.stack_scalar(val, batch_size=batch_size)

        if aggregation is None:
            return val

        agg_dict = {'tensor': {'mean': torch.mean, 'sum': torch.sum, 'max': torch.max, 'min': torch.min, 'std': torch.std,
                    'median': torch.median, 'var': torch.var},
                    'numpy': {'mean': np.mean, 'sum': np.sum, 'max': np.max, 'min': np.min, 'std': np.std,
                                'median': np.median, 'var': np.var},
                    }

        val = recursive_flatten(val, flat_array=True, tolist=False)
        val_type = check_type(val)
        if val_type.major == 'scalar':
            val = float(val)
        elif val_type.minor == 'pandas':
            val = val.values
            val = agg_dict['numpy'][aggregation](val)
        elif val_type.minor == 'tensor':
            val = agg_dict['tensor'][aggregation](val)
        else:
            val = agg_dict['numpy'][aggregation](val)

        val = squeeze_scalar(val)

        return val

    def get_scalar(self, name, subset=None, aggregate=False, stack=False, index=None):
        if subset is None:
            subset = self.subset_context
        k = name if subset is None else f'{subset}/{name}'

        if k in self.scalar:
            v = self.scalar[k]
            if aggregate:
                agg = self.scalar_aggregation.get(f'{subset}/{name}', 'mean')
                v = self.aggregate_scalar(v, agg, batch_size=self.batch_size_context)
            elif index is not None:
                v = v[index]
            elif stack:
                v = self.stack_scalar(v, batch_size=self.batch_size_context)
            return v
        return None

    def get_scalars(self, name, subset=None, aggregate=False):
        if subset is None:
            subset = self.subset_context

        key = name if subset is None else f'{subset}/{name}'
        if key in self.scalars:
            if aggregate:
                agg = self.scalar_aggregation.get(key, 'mean')
                val_dict = self.scalars[key]
                val_dict_agg = {}

                for k, val in val_dict.items():
                    val_dict_agg[k] = self.aggregate_scalar(val, agg, batch_size=self.batch_size_context)

                return val_dict_agg
            return self.scalars[key]
        return None

    def get_data(self, name, data_type=None, subset=None):
        if subset is None:
            subset = self.subset_context
        key = name if subset is None else f'{subset}/{name}'
        if key in self.data[data_type]:
            return self.data[data_type][key]
        return None

    def get_buffer(self, name, subset=None):
        if subset is None:
            subset = self.subset_context
        key = name if subset is None else f'{subset}/{name}'
        if key in self.buffer:
            return self.buffer[key]
        return None

    def get_image(self, name, subset=None):
        return self.get_data(name, data_type='image', subset=subset)

    def get_images(self, name, subset=None):
        return self.get_data(name, data_type='images', subset=subset)

    def get_histogram(self, name, subset=None):
        return self.get_data(name, data_type='histogram', subset=subset)

    def get_figure(self, name, subset=None):
        return self.get_data(name, data_type='figure', subset=subset)

    def get_audio(self, name, subset=None):
        return self.get_data(name, data_type='audio', subset=subset)

    def get_video(self, name, subset=None):
        return self.get_data(name, data_type='video', subset=subset)

    def get_text(self, name, subset=None):
        return self.get_data(name, data_type='text', subset=subset)

    def get_embedding(self, name, subset=None):
        return self.get_data(name, data_type='embedding', subset=subset)

    def get_mesh(self, name, subset=None):
        return self.get_data(name, data_type='mesh', subset=subset)

    def get_pr_curve(self, name, subset=None):
        return self.get_data(name, data_type='pr_curve', subset=subset)

    def report_scalar(self, name, val, subset=None, aggregation=None, append=None, **kwargs):

        if subset is None:
            subset = self.subset_context

        key = name if subset is None else f'{subset}/{name}'
        if append is None:
            append = self.state == 'in_epoch'

        val = self.detach_scalar(val)

        if append:
            self.scalar[key].append(val)
        else:
            self.scalar[key] = val

        kwargs['global_step'] = self.epoch
        self.scalar_kwargs[key] = kwargs

        if name not in self.subsets_keys[subset]['scalar']:
            self.subsets_keys[subset]['scalar'].append(name)

        if aggregation is None:
            aggregation = 'mean'
        self.scalar_aggregation[key] = aggregation

    def report_scalars(self, name, val_dict, subset=None, aggregation=None, append=None, **kwargs):

        if subset is None:
            subset = self.subset_context

        key = name if subset is None else f'{subset}/{name}'
        if append is None:
            append = self.state == 'in_epoch'

        for k, val in val_dict.items():

            val = self.detach_scalar(val)
            if append:
                self.scalars[key][k].append(val)
            else:
                self.scalars[key][k] = val

        kwargs['global_step'] = self.epoch
        self.scalar_kwargs[key] = kwargs

        if name not in self.subsets_keys[subset]['scalars']:
            self.subsets_keys[subset]['scalars'].append(name)

        if aggregation is None:
            aggregation = 'mean'
        self.scalars_aggregation[key] = aggregation

    def report_data(self, name, val, subset=None, data_type=None, **kwargs):

        if subset is None:
            subset = self.subset_context

        key = name if subset is None else f'{subset}/{name}'
        if data_type is None:
            data_type = 'other'

        self.aux[data_type][key] = val

        if name not in self.subsets_keys[subset][data_type]:
            self.subsets_keys[subset][data_type].append(name)

        kwargs['global_step'] = self.epoch
        self.aux_kwargs[data_type][key] = kwargs

    def report_histogram(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'histogram', **kwargs)

    def report_image(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'image', **kwargs)

    def report_images(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'images', **kwargs)

    def report_figure(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'figure', **kwargs)

    def report_video(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'video', **kwargs)

    def report_audio(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'audio', **kwargs)

    def report_text(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'text', **kwargs)

    def report_embedding(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'embedding', **kwargs)

    def report_mesh(self, name, val, subset=None, **kwargs):
        self.report_data(name, val, subset, 'mesh', **kwargs)

    def report_pr_curve(self, name, labels, predictions, subset=None, **kwargs):
        kwargs['predictions'] = predictions
        self.report_data(name, labels, subset, 'pr_curve', **kwargs)

    def add_buffer(self, name, val, subset=None):
        if subset is None:
            subset = self.subset_context

        if name not in self.subsets_keys[subset]['buffer']:
            self.subsets_keys[subset]['buffer'].append(name)

        if subset is None:
            self.buffer[name].append(val)
        else:
            self.buffer[subset][name].append(val)



import argparse
import os
import math
from pathlib import Path
from .core import BeamConfig, BeamParam
from .deepspeed import DeepspeedConfig
from .utils import _beam_arguments


class CatboostConfig(BeamConfig):
    # catboost
    parameters = [
        BeamParam('cb_ranker', bool, False, 'Whether to use catboost ranker instead of regression', tags='model'),
        BeamParam('cb_n_estimators', int, 1000, 'The number of trees in the catboost model', tags='tune'),
        BeamParam('cb_l2_leaf_reg', float, 1e-4, 'The L2 regularization for the catboost model', tags='tune'),
        BeamParam('cb_border_count', int, 128, 'The border count for the catboost model', tags='tune'),
        BeamParam('cb_depth', int, 14, 'The depth of the trees in the catboost model', tags='tune'),
        BeamParam('cb_random_strength', float, .5, 'The random strength for the catboost model', tags='tune'),
    ]


class NNModelConfig(BeamConfig):
    parameters = [
        BeamParam('init', str, 'ortho', 'Initialization method [ortho|N02|xavier|]', tags='tune'),
    ]


class SchedulerConfig(BeamConfig):
    
    parameters = [
        BeamParam('scheduler_steps', str, 'epoch', 'When to apply schedulers steps [epoch|iteration|none]: '
                                                   'each epoch or each iteration. '
                                                   'Use none to avoid scheduler steps or to use your own custom steps policy'),
    
        BeamParam('scheduler', str, None, 'Build BeamScheduler. Supported schedulers: '
                                          '[one_cycle,reduce_on_plateau,cosine_annealing]', tags='tune'),
    
        BeamParam('cycle_base_momentum', float, .85, 'The base momentum in one-cycle optimizer', tags='tune'),
        BeamParam('cawr_t0', int, 10, ' Number of iterations for the first restart in CosineAnnealingWarmRestarts scheduler',
                    tags='tune'),
        BeamParam('cawr_tmult', int, 1, ' A factor increases Ti after a restart in CosineAnnealingWarmRestarts scheduler',
                    tags='tune'),
        BeamParam('scheduler_factor', float, math.sqrt(.1), 'The factor to reduce lr in schedulers such as ReduceOnPlateau',
                    tags='tune'),
        BeamParam('scheduler_patience', int, 10, 'Patience for the ReduceOnPlateau scheduler', tags='tune'),
        BeamParam('scheduler_warmup', float, 5, 'Scheduler\'s warmup factor (in epochs)', tags='tune'),
        BeamParam('cycle_max_momentum', float, .95, 'The maximum momentum in one-cycle scheduler', tags='tune'),
    ]


class DatasetConfig(BeamConfig):

    parameters = [
        BeamParam('split_dataset_seed', int, 5782, 'Seed dataset split (set to zero to get random split)'),
        BeamParam('test_size', float, .2, 'Test set percentage'),
        BeamParam('validation_size', float, .2, 'Validation set percentage'),
    ]


class SamplerConfig(BeamConfig):

    parameters = [
        BeamParam('oversampling_factor', float, .0, 'A factor [0, 1] that controls how much to oversample where '
                                                    '0-no oversampling and 1-full oversampling. Set 0 for no oversampling',
                  tags='tune'),
        BeamParam('expansion_size', int, int(1e7), 'largest expanded index size for oversampling'),
        BeamParam('dynamic_sampler', bool, False, 'Whether to use a dynamic sampler (mainly for rl/optimization)'),
        BeamParam('buffer_size', int, None, 'Maximal Dataset size in dynamic problems', tags='tune'),
        BeamParam('probs_normalization', str, 'sum', 'Sampler\'s probabilities normalization method [sum/softmax]'),
        BeamParam('sample_size', int, 100000, 'Periodic sample size for the dynamic sampler'),
    ]


class OptimizerConfig(BeamConfig):

    parameters = [
        BeamParam('weight_decay', float, 0., 'L2 regularization coefficient for dense optimizers', tags='tune'),
        BeamParam('eps', float, 1e-4, 'Adam\'s epsilon parameter', tags='tune'),
        BeamParam('momentum', float, .9, 'The momentum and Adam\'s β1 parameter', tags='tune'),
        BeamParam('beta2', float, .999, 'Adam\'s β2 parameter', tags='tune'),
        BeamParam('clip_gradient', float, 0., 'Clip Gradient L2 norm', tags='tune'),
        BeamParam('accumulate', int, 1, 'Accumulate gradients for this number of backward iterations', tags='tune'),
    ]


class SWAConfig(BeamConfig):

    parameters = [
        BeamParam('swa', float, None, 'SWA period. If float it is a fraction of the total number of epochs. '
                                      'If integer, it is the number of SWA epochs.'),
        BeamParam('swa_lr', float, 0.05, 'The SWA learning rate', tags='tune'),
        BeamParam('swa_anneal_epochs', int, 10, 'The SWA lr annealing period', tags='tune'),
    ]


class FederatedTrainingConfig(BeamConfig):

    parameters = [
        BeamParam('mp_ip', str, 'localhost', 'IP to be used for multiprocessing'),
        BeamParam('mp_port', str, None, 'Port to be used for multiprocessing'),
        BeamParam('n_gpus', int, 1, 'Number of parallel gpu workers. Set <=1 for single process'),
        BeamParam('n_cpus_per_worker', int, 6, 'Number of cpus to use in each worker'),
        BeamParam('n_gpus_per_worker', int, 1, 'Number of gpus to use in each worker'),
        BeamParam('distributed_backend', str, 'nccl', 'The distributed backend to use. '
                                                    'Supported backends: [nccl, gloo, mpi]'),
        BeamParam('mp_context', str, 'spawn', 'The multiprocessing context to use'),
        BeamParam('kv_store', str, 'tcp', 'The key-value store to use [tcp|file|hash]'),
        BeamParam('kv_store_path', str, None, 'The path to the key-value store file'),
        BeamParam('kv_store_timeout', float, 300., 'The timeout for the key-value store'),
        BeamParam('kv_store_port', str, None, 'The port to use for the key-value store'),
        BeamParam('federated_runner', bool, False, 'Use the new federated runner for distributed training'),

    ]


class AccelerateConfig(DeepspeedConfig, FederatedTrainingConfig):
    # accelerate parameters
    # based on https://huggingface.co/docs/accelerate/v0.24.0/en/package_reference/accelerator#accelerate.Accelerator

    # boolean_feature(parser, "deepspeed-dataloader", False,
    #                 "Use optimized deepspeed dataloader instead of native pytorch dataloader")

    parameters = [
        BeamParam('device_placement', bool, False, 'Whether or not the accelerator should put objects on device'),
        BeamParam('split_batches', bool, False, 'Whether or not the accelerator should split the batches '
                                                'yielded by the dataloaders across the devices'),
    ]


class NNCompilerConfig(BeamConfig):

    """
    For torch compile see:
    https://pytorch.org/docs/stable/generated/torch.compile.html

    For torch jit see:
    https://pytorch.org/docs/stable/generated/torch.jit.trace.html
    """

    parameters = [
        BeamParam('compile_fullgraph', bool, False, 'Whether it is ok to break model into several subgraphs'),
        BeamParam('compile_dynamic', bool, None, 'Use dynamic shape tracing. When this is True, '
                                                  'we will up-front attempt to generate a kernel that is as dynamic as '
                                                  'possible to avoid recompilations when sizes change.'),
        BeamParam('compile_backend', str, 'inductor', 'The backend to use for compilation [inductor|torch]'),
        BeamParam('compile_mode', str, 'default', '[default|reduce-overhead|max-autotune|max-autotune-no-cudagraphs],'
                                                  ' see https://pytorch.org/docs/stable/generated/torch.compile.html'),
        BeamParam('compile_options', dict, None, 'Additional options for the compiler'),


        BeamParam('jit_check_trace', bool, True, 'Check if the same inputs run through traced code produce the same '
                                                 'outputs'),
        BeamParam('jit_check_inputs', list, None, 'A list of tuples of input arguments that should be used to check the '
                                                  'trace against what is expected.'),
        BeamParam('jit_check_tolerance', float, 1e-5, 'Floating-point comparison tolerance to use in the checker '
                                                      'procedure.'),
        BeamParam('jit_strict', bool, True, 'Only turn this off when you want the tracer to record your mutable container '
                                            'types'),
    ]


class NNTrainingConfig(NNModelConfig, SchedulerConfig, AccelerateConfig,
                       SWAConfig, OptimizerConfig, DatasetConfig, SamplerConfig, NNCompilerConfig):

    parameters = [
        BeamParam('objective', str, 'objective', 'A single objective to apply hyperparameter optimization or '
                                                 'ReduceLROnPlateau scheduling. '
                                                 'By default we consider maximization of the objective (e.g. accuracy) '
                                                 'You can override this behavior by overriding the Algorithm.report method.'),

        BeamParam('objective_mode', str, None, 'Set [min/max] to minimize/maximize the objective. '
                                               'By default objectives that contain the words "loss/error/mse" a are minimized and '
                                               'other objectives are maximized. You can override this behavior by setting this flag.'),

        BeamParam('objective_to_report', str, 'best', 'Which objective to report in HPO run [best|last]'),

        BeamParam('scale_epoch_by_batch_size', bool, True,
                  'When True: epoch length corresponds to the number of examples sampled from the dataset in each epoch '
                  'When False: epoch length corresponds to the number of forward passes in each epoch'),

        BeamParam('model_dtype', str, 'float32', 'dtype, both for automatic mixed precision and accelerate. '
                                                 'Supported dtypes: [float32, float16, bfloat16]', tags=['tune', 'model']),
        BeamParam('total_steps', int, int(1e6), 'Total number of environment steps', tags='tune'),
        BeamParam('epoch_length', int, None, 'Length of train+eval epochs '
                                             '(if None - it is taken from epoch_length_train/epoch_length_eval arguments)',
                  tags='tune'),
        BeamParam('train_on_tail', bool, False, 'Should the last (incomplete) batch be included in the training epoch '
                                                '(useful for small datasets but not time efficient)'),
        BeamParam('epoch_length_train', int, None, 'Length of each epoch (if None - it is the dataset[train] size)', tags='tune'),
        BeamParam('epoch_length_eval', int, None, 'Length of each evaluation epoch (if None - it is the dataset[validation] size)',
                    tags='tune'),
        BeamParam('n_epochs', int, None, 'Number of epochs, if None, '
                                         'it uses the total steps to determine the number of iterations', tags='tune'),
        BeamParam('batch_size', int, 256, 'Batch Size', tags='tune'),
        BeamParam('batch_size_train', int, None, 'Batch Size for training iterations', tags='tune'),
        BeamParam('batch_size_eval', int, None, 'Batch Size for testing/evaluation iterations', tags='tune'),
        BeamParam('reduction', str, 'sum', 'whether to sum loss elements or average them [sum|mean|mean_batch|sqrt|mean_sqrt]',
                    tags='tune'),
        BeamParam('lr_dense', float, 1e-3, 'learning rate for dense optimizers', tags='tune'),
        BeamParam('lr_sparse', float, 1e-2, 'learning rate for sparse optimizers', tags='tune'),
        BeamParam('stop_at', float, None, 'Early stopping when objective >= stop_at', tags='tune'),
        BeamParam('early_stopping_patience', int, None, 'Early stopping patience in epochs, '
                                                     'stop when current_epoch - best_epoch >= early_stopping_patience',
                    tags='tune'),
    ]


class KeysConfig(BeamConfig):

    parameters = [
        BeamParam('COMET_API_KEY', str, None, 'The comet.ml api key to use for logging'),
        BeamParam('AWS_ACCESS_KEY_ID', str, None, 'The aws access key to use for S3 connections'),
        BeamParam('AWS_SECRET_ACCESS_KEY', str, None, 'The aws private key to use for S3 connections'),
        BeamParam('SSH_PRIVATE_KEY', str, None, 'The ssh secret key to use for ssh connections'),
        BeamParam('OPENAI_API_KEY', str, None, 'The openai api key to use for openai connections'),
        BeamParam('BEAM_USERNAME', str, None, 'The beam username to use for connections like smb/ftp/ssh etc'),
        BeamParam('BEAM_PASSWORD', str, None, 'The beam password to use for connections like smb/ftp/ssh etc'),
    ]


class DDPConfig(BeamConfig):

    parameters = [
        BeamParam('find_unused_parameters', bool, False, 'For DDP applications: allows running backward on '
                                                         'a subgraph of the model. introduces extra overheads, '
                                                         'so applications should only set find_unused_parameters '
                                                         'to True when necessary'),

        BeamParam('broadcast_buffers', bool, True, 'For DDP applications: Flag that enables syncing (broadcasting) '
                                                   'buffers of the module at beginning of the forward function.'),

        BeamParam('nvlink', bool, False, 'For DDP applications: whether nvlink is available for faster communication'),
    ]


class BeamProjectConfig(BeamConfig):
    parameters = [
        BeamParam('project_name', str, 'beam', 'The name of the beam project'),
        BeamParam('algorithm', str, 'Algorithm', 'algorithm name'),
        BeamParam('identifier', str, 'debug', 'The name of the model to use'),
        BeamParam('logs_path', str, os.path.join(os.path.expanduser('~'), 'beam_projects', 'experiments'),
                  'Root directory for Logs and results'),
        BeamParam('data_path', str, os.path.join(os.path.expanduser('~'), 'beam_projects', 'data'),
                  'Where the dataset is located'),
        BeamParam('config_file', str, str(Path.home().joinpath('conf.pkl'), ),
                  'The beam config file to use with secret keys'),
        BeamParam('verbosity', str, 'info', 'The verbosity level [debug|info|warning|error]'),
    ]


class ExperimentConfig(BeamProjectConfig, NNTrainingConfig, DDPConfig, KeysConfig):
    '''

        Arguments

            global parameters

            These parameters are responsible for which experiment to load or to generate:
            the name of the experiment is <alg>_<identifier>_exp_<num>_<time>
            The possible configurations:
            reload = False, override = True: always overrides last experiment (default configuration)
            reload = False, override = False: always append experiment to the list (increment experiment num)
            reload = True, resume = -1: resume to the last experiment
            reload = True, resume = <n>: resume to the <n> experiment

    '''

    parameters = [

        BeamParam('llm', str, None, 'URI of a Large Language Model to be used in the experiment.'),
        BeamParam('reload', bool, False, 'Load saved model'),
        BeamParam('resume', int, -1, 'Resume experiment number, set -1 for last experiment: active when reload=True'),
        BeamParam('override', bool, False, 'Override last experiment: active when reload=False'),
        BeamParam('reload_checkpoint', str, 'best', 'Which checkpoint to reload [best|last|<epoch>]'),

        BeamParam('cpu_workers', int, 0, 'How many CPUs will be used for the data loading'),
        BeamParam('data_fetch_timeout', float, 0., 'Timeout for the dataloader fetching. '
                                                   'set to 0 for no timeout.'),
        BeamParam('device', str, '0', 'GPU Number or cpu/cuda string'),
        BeamParam('device_list', list, None,
                  'Set GPU priority for parallel execution e.g. --device-list 2 1 3 will use GPUs 2 and 1 '
                  'when passing --n-gpus=2 and will use GPUs 2 1 3 when passing --n-gpus=3. '
                  'If None, will use an ascending order starting from the GPU passed in the --device parameter. '
                  'e.g. when --device=1 will use GPUs 1,2,3,4 when --n-gpus=4'),

        BeamParam('tensorboard', bool, True, 'Log results to tensorboard'),
        BeamParam('mlflow', bool, False, 'Log results to MLFLOW serve'),
        BeamParam('lognet', bool, True, 'Log  networks parameters'),
        BeamParam('deterministic', bool, False, 'Use deterministic pytorch optimization for reproducability '
                                                'when enabling non-deterministic behavior, it sets '
                                                'torch.backends.cudnn.benchmark = True which '
                                                'accelerates the computation'),

        BeamParam('scalene', bool, False, 'Profile the experiment with the Scalene python profiler'),
        BeamParam('safetensors', bool, False, 'Save tensors in safetensors format instead of native torch'),
        BeamParam('store_initial_weights', bool, False, 'Store the network\'s initial weights'),
        BeamParam('store_init_args', bool, True, 'Store the algorithm init args/kwargs for better reloading'),
        BeamParam('copy_code', bool, True, 'Copy the code directory into the experiment directory'),
        BeamParam('restart_epochs_count', bool, True, 'When reloading an algorithm, restart counting epochs from zero '
                                                      '(with respect to schedulers and swa training)', tags='tune'),
        BeamParam('seed', int, 0, 'Seed for reproducability (zero is saved for random seed)'),
        BeamParam('train_timeout', int, None, 'Timeout for the training in seconds. Set to None for no timeout',
                  tags='tune'),

        # results printing and visualization

        BeamParam('print_results', bool, True, 'Print results after each epoch to screen'),
        BeamParam('visualize_weights', bool, False, 'Visualize network weights on tensorboard'),
        BeamParam('enable_tqdm', bool, True, 'Print tqdm progress bar when training'),
        BeamParam('visualize_results_log_base', int, 10, 'log base for the logarithmic based results visualization'),
        BeamParam('tqdm_threshold', float, 10., 'Minimal expected epoch time to print tqdm bar '
                                                'set 0 to ignore and determine tqdm bar with tqdm-enable flag'),
        BeamParam('tqdm_stats', float, 1., 'Take this period to calculate the expected epoch time'),

        BeamParam('visualize_results', str, 'yes', 'when to visualize results on tensorboard '
                                                   '[yes|no|logscale|best|never|final]'),
        BeamParam('store_results', str, 'logscale', 'when to store results to pickle files '
                                                    '[yes|no|logscale|best|never|final]'),
        BeamParam('store_networks', str, 'best/last', 'when to store network weights to the log directory '
                                         '[yes|no|logscale|best|all_bests|never|final|last|best/last]'),

        BeamParam('comet', bool, False, 'Whether to use comet.ml for logging'),
        BeamParam('git_directory', str, None, 'The git directory to use for comet.ml logging'),
        BeamParam('comet_workspace', str, None, 'The comet.ml workspace to use for logging'),

        BeamParam('mlflow_url', str, None, 'The url of the mlflow serve to use for logging. '
                                           'If None, mlflow will log to $MLFLOW_TRACKING_URI'),

        BeamParam('training_framework', str, 'torch', 'Chose between [torch|amp|accelerate|deepspeed]'),
        BeamParam('compile_train', bool, False,
                  'Apply torch.compile to optimize the inner_train function to speed up training. '
                  'To use this feature, you must override and use the alg.inner_train function '
                  'in your alg.train_iteration function'),
        BeamParam('compile_network', bool, False,
                  'Apply torch.compile to optimize the network forward function to speed up training.'),

        # possible combinations for single gpu:
        # 1. torch
        # 2. amp
        # 3. accelerate
        # 4. native deepspeed

        # possible combinations for multiple gpus:
        # 1. torch + ddp
        # 2. amp + ddp
        # 3. accelerate + deepspeed
        # 4. native deepspeed
    ]


class TransformerConfig(BeamConfig):
    # transformer arguments

    parameters = [
        BeamParam('mp_method', str, 'joblib', 'The multiprocessing method to use'),
        BeamParam('n_chunks', int, None, 'The number of chunks to split the dataset'),
        BeamParam('name', str, None, 'The name of the dataset', tags='tune'),
        BeamParam('store_path', str, None, 'The path to store the results'),
        BeamParam('partition', str, None, 'The partition to use for splitting the dataset'),
        BeamParam('chunksize', int, None, 'The chunksize to use for splitting the dataset'),
        BeamParam('squeeze', bool, True, 'Whether to squeeze the results'),
        BeamParam('reduce', bool, True, 'Whether to reduce and collate the results'),
        BeamParam('reduce_dim', int, None, 'The dimension to reduce the results'),
        BeamParam('transform_strategy', str, None, 'The transform strategy to use can be [CC|CS|SC|SS]'),
        BeamParam('split_by', str, 'keys', 'The split strategy to use can be [keys|index|columns]'),
        BeamParam('store_suffix', str, None, 'The suffix to add to the stored file'),
    ]


class UniversalConfig(ExperimentConfig, TransformerConfig, CatboostConfig):
    pass


def beam_arguments(*args, return_defaults=False, return_tags=False, **kwargs):

    if len(args) and type(args[0]) == argparse.ArgumentParser:
        pr = args[0]
        args = args[1:]
    else:
        pr = UniversalConfig()

    args = [pr] + list(args)
    return _beam_arguments(*args, return_defaults=return_defaults, return_tags=return_tags, **kwargs)



from .utils import to_dict, get_beam_llm, print_beam_hyperparameters
from .configurations import *
from .core import BeamConfig, BeamParam
from .deepspeed import DeepspeedConfig, deepspeed_config_generator



from .k8s import BeamK8S
from .deploy import BeamDeploy, ServiceConfig, StorageConfig, UserIdmConfig, SecurityContextConfig, MemoryStorageConfig, RayPortsConfig
from .pod import BeamPod
from .units import K8SUnits




from dataclasses import dataclass

from prefect import flow, task
from prefect.task_runners import SequentialTaskRunner
import datetime

from ..path import beam_path
from ..utils import beam_hash
from ..core import Processor
from ..transformer import Transformer


@dataclass
class CronSchedule:
    hour: int = 0           # Hour of the day (0-23)
    minute: int = 0         # Minute of the hour (0-59)
    day_of_month: str = '*' # Day of the month (1-31) or '*' for every day of the month
    month: str = '*'        # Month of the year (1-12) or '*' for every month
    day_of_week: str = '*'  # Day of the week (0-6, where 0 is Sunday) or '*' for every day of the week

    def __str__(self):
        """
        Return a cron expression string based on the schedule parameters.
        """
        return f"{self.minute} {self.hour} {self.day_of_month} {self.month} {self.day_of_week}"

    @property
    def str(self):
        return self.__str__()


class BeamFlow(Processor):

    def __init__(self, obj, *args, description=None, log_prints=True, retries=None, retry_delay_seconds=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = obj
        self.description = description
        self.log_prints = log_prints
        self.retries = retries
        self.retry_delay_seconds = retry_delay_seconds
        self.name = obj.name if hasattr(obj, 'name') else self.name
        self.is_transformer = isinstance(obj, Transformer)
        self.run = task(description=self.description, name=self.name)(self._run)
        self.flow = flow(name=self.name, log_prints=self.log_prints,
                         retries=self.retries, retry_delay_seconds=self.retry_delay_seconds)(self._flow)
        self.monitored_paths = {}

    def schedule(self, method, hour: int = 0, minute: int = 0, day_of_month: str = '*',
                 month: str = '*', day_of_week: str = '*', cron_schedule: CronSchedule = None):

        if cron_schedule is None:
            cron_schedule = CronSchedule(hour=hour, minute=minute, day_of_month=day_of_month,
                                         month=month, day_of_week=day_of_week)

        self.flow.serve(method, schedule=cron_schedule.str)

    def monitor_path(self, path, lazy=False):
        path = beam_path(path)
        walk = path.walk()
        walk_hash = beam_hash(walk)

        path_uri = path.as_uri()
        if path_uri not in self.monitored_paths:
            self.monitored_paths[walk_hash] = path
            return not lazy
        elif self.monitored_paths[path_uri] != walk_hash:
            self.monitored_paths[path_uri] = walk_hash
            return True
        return False

    def _run(self, method, *args, **kwargs):

        method = getattr(self.obj, method)
        result = method(*args, **kwargs)
        return result

    def _flow(self, method, args, kwargs):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        return self.run(method, args=args, kwargs=kwargs)


def beam_flow(obj, method, *args, **kwargs):

    bf = BeamFlow(obj)

    @flow(name=obj.name, log_prints=True)
    def _beam_flow(obj, method, args=None, kwargs=None):

        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        return bf.run(obj, method, args, kwargs)

    _beam_flow(obj=obj, method=method, args=args, kwargs=kwargs)





# @task
# def execute_method(obj, method, *args, **kwargs):
#     scheduler = BeamFlow(obj)
#     return scheduler.run(method, *args, **kwargs)
#
#
# def beam_scheduler(obj, method, *args, **kwargs):
#     @flow(task_runner=SequentialTaskRunner(), name=obj.name, log_prints=True)
#     def beam_scheduler_flow(obj, method, args=None, kwargs=None):
#
#         if args is None:
#             args = ()
#         if kwargs is None:
#             kwargs = {}
#
#         return execute_method(obj, method, args, kwargs)
#
#     beam_scheduler_flow.serve(obj=obj, method=method, args=args, kwargs=kwargs)


# Example usage
# flow.run(obj=your_object, method='your_method', args=your_args, kwargs=your_kwargs)




# def schedule(obj, method, ...scheduling args...):
#     scheduler = BeamScheduler(obj)
#     ...run this with prefect scheduler...




import torch
from ..utils import as_numpy, as_tensor
from ..data import BeamData
from ..dataset import UniversalDataset
import numpy as np
import pandas as pd
from ..logger import beam_logger as logger


class TabularDataset(UniversalDataset):

    def __init__(self, hparams):

        bd = BeamData.from_path(hparams.data_path)
        dataset = bd[hparams.dataset_name].cached()
        info = dataset['info'].values
        self.task_type = info['task_type']

        x_train = dataset['N_train'].values
        x_val = dataset['N_val'].values
        x_test = dataset['N_test'].values

        if np.isnan(x_train).any() or np.isnan(x_val).any() or np.isnan(x_test).any():
            logger.warning('NaN values in the data, replacing with 0')
            x_train = np.nan_to_num(x_train)
            x_val = np.nan_to_num(x_val)
            x_test = np.nan_to_num(x_test)

        y_train = dataset['y_train'].values

        self.numerical_features, self.cat_features = self.get_numerical_and_categorical(x_train, y_train)

        x_train_num = x_train[:, self.numerical_features]
        x_train_cat = x_train[:, self.cat_features].astype(np.int64)

        x_val_num = x_val[:, self.numerical_features]
        x_val_cat = x_val[:, self.cat_features].astype(np.int64)

        x_test_num = x_test[:, self.numerical_features]
        x_test_cat = x_test[:, self.cat_features].astype(np.int64)

        if hparams.oh_to_cat:
            self.oh_categories = self.one_hot_to_categorical(x_train_cat)

            x_val_cat = np.stack([x_val_cat.T[self.oh_categories == c].argmax(axis=0)
                                  for c in np.unique(self.oh_categories)], axis=1)
            x_train_cat = np.stack([x_train_cat.T[self.oh_categories == c].argmax(axis=0)
                                    for c in np.unique(self.oh_categories)], axis=1)
            x_test_cat = np.stack([x_test_cat.T[self.oh_categories == c].argmax(axis=0)
                                   for c in np.unique(self.oh_categories)], axis=1)

        if info['n_cat_features'] > 0:

            d = dataset['C_trainval'].values
            factors = [pd.factorize(d[:, i])[1] for i in range(d.shape[1])]

            d = dataset['C_train'].values
            x_train_cat_aux = np.stack([pd.Categorical(d[:, i], categories=f).codes
                                        for i, f in enumerate(factors)], axis=1).astype(np.int64)
            d = dataset['C_val'].values
            x_val_cat_aux = np.stack([pd.Categorical(d[:, i], categories=f).codes
                                      for i, f in enumerate(factors)], axis=1).astype(np.int64)
            d = dataset['C_test'].values
            x_test_cat_aux = np.stack([pd.Categorical(d[:, i], categories=f).codes
                                        for i, f in enumerate(factors)], axis=1).astype(np.int64)

            # plus 1 for nan values
            x_train_cat = np.concatenate([x_train_cat, x_train_cat_aux+1], axis=1)
            x_val_cat = np.concatenate([x_val_cat, x_val_cat_aux+1], axis=1)
            x_test_cat = np.concatenate([x_test_cat, x_test_cat_aux+1], axis=1)

        if hparams.scaler == 'robust':
            from sklearn.preprocessing import RobustScaler
            self.scaler = RobustScaler()
        elif hparams.scaler == 'quantile':
            from sklearn.preprocessing import QuantileTransformer
            self.scaler = QuantileTransformer(n_quantiles=1000, subsample=100000, random_state=hparams.seed)
        else:
            raise ValueError('Unknown scaler')

        self.scaler.fit(x_train_num)

        x_train_num_scaled = torch.FloatTensor(self.scaler.transform(x_train_num))
        x_val_num_scaled = torch.FloatTensor(self.scaler.transform(x_val_num))
        x_test_num_scaled = torch.FloatTensor(self.scaler.transform(x_test_num))

        # save these tables for catboost training
        self.x_train_num_scaled = as_numpy(x_train_num_scaled)
        self.x_val_num_scaled = as_numpy(x_val_num_scaled)
        self.x_test_num_scaled = as_numpy(x_test_num_scaled)
        self.x_train_cat = x_train_cat
        self.x_val_cat = x_val_cat
        self.x_test_cat = x_test_cat
        self.y_train = dataset['y_train'].values
        self.y_val = dataset['y_val'].values
        self.y_test = dataset['y_test'].values

        self.y_mu = None
        self.y_sigma = None
        if self.task_type == 'regression':
            y_train = torch.FloatTensor(dataset['y_train'].values)
            y_val = torch.FloatTensor(dataset['y_val'].values)
            y_test = torch.FloatTensor(dataset['y_test'].values)

            mu = y_train.mean(dim=0, keepdim=True)
            sigma = y_train.std(dim=0, keepdim=True)

            self.y_mu = float(mu)
            self.y_sigma = float(sigma)

            y_train = (y_train - mu) / (sigma + 1e-8)
            y_val = (y_val - mu) / (sigma + 1e-8)
            y_test = (y_test - mu) / (sigma + 1e-8)

        else:
            y_train = torch.LongTensor(dataset['y_train'].values)
            y_val = torch.LongTensor(dataset['y_val'].values)
            y_test = torch.LongTensor(dataset['y_test'].values)

        n_quantiles = hparams.n_quantiles
        x_train_num_quantized = (x_train_num_scaled * n_quantiles).long()
        x_val_num_quantized = (x_val_num_scaled * n_quantiles).long()
        x_test_num_quantized = (x_test_num_scaled * n_quantiles).long()

        x_train_num_fractional = x_train_num_scaled * n_quantiles - x_train_num_quantized.float()
        x_val_num_fractional = x_val_num_scaled * n_quantiles - x_val_num_quantized.float()
        x_test_num_fractional = x_test_num_scaled * n_quantiles - x_test_num_quantized.float()

        self.cat_mask = torch.cat([torch.ones(x_train_num_quantized.shape[-1]), torch.zeros(x_train_cat.shape[-1])])

        x_train_mixed = torch.cat([x_train_num_quantized, as_tensor(x_train_cat)], dim=1)
        x_val_mixed = torch.cat([x_val_num_quantized, as_tensor(x_val_cat)], dim=1)
        x_test_mixed = torch.cat([x_test_num_quantized, as_tensor(x_test_cat)], dim=1)

        self.n_tokens = torch.stack([xi.max(dim=0).values
                                     for xi in [x_train_mixed, x_val_mixed, x_test_mixed]]).max(dim=0).values + 1

        x_train_frac = torch.cat([x_train_num_fractional, torch.zeros(x_train_cat.shape)], dim=1)
        x_val_frac = torch.cat([x_val_num_fractional, torch.zeros(x_val_cat.shape)], dim=1)
        x_test_frac = torch.cat([x_test_num_fractional, torch.zeros(x_test_cat.shape)], dim=1)

        x = torch.cat([x_train_mixed, x_val_mixed, x_test_mixed], dim=0)
        x_frac = torch.cat([x_train_frac, x_val_frac, x_test_frac], dim=0)
        y = torch.cat([y_train, y_val, y_test], dim=0)

        device = None
        if hparams.store_data_on_device:
            device = hparams.device

        super().__init__(x=x, x_frac=x_frac, label=y, device=device)

        if self.task_type == 'regression':
            self.n_classes = 1
        else:
            self.n_classes = self.label.max() + 1

        # logger.warning("Check x-max consistency at dataset:")
        # logger.warning(f"n-tokens: {self.n_tokens}")
        # logger.warning(self.n_tokens < x.max(dim=0).values)
        self.split(validation=len(x_train_mixed) + np.arange(len(x_val_mixed)),
                           test=len(x_train_mixed) + len(x_val_mixed) + np.arange(len(x_test_mixed)))

    @staticmethod
    def get_numerical_and_categorical(x, y=None):
        """
        @param x: input data
        @return: numerical and categorical features
        """
        import deepchecks as dch
        dataset = dch.tabular.Dataset(x, label=y)

        return dataset.numerical_features, dataset.cat_features

    @staticmethod
    def one_hot_to_categorical(x):
        """
        @param x: one-hot encoded categorical features
        @return: mapping from one-hot to categorical
        """
        return x.cumsum(axis=1).max(axis=0)




from .resource import beam_worker, beam_dispatcher_server, beam_dispatcher, async_client

from .celery_dispatcher import CeleryDispatcher
from .celery_worker import CeleryWorker

from .ray_dispatcher import RayDispatcher, RayCluster
from .thread_dispatcher import ThreadedDispatcher

from .async_client import AsyncClient
# from .async_server import AsyncRayServer, AsyncCeleryServer



from ..utils import (collate_chunks, recursive_chunks, iter_container,
                        build_container_from_tupled_keys, is_empty, )
from ..parallel import BeamParallel, BeamTask
from ..data import BeamData
from ..path import beam_path
from ..utils import tqdm_beam as tqdm
from ..logger import beam_logger as logger
from ..core.processor import Processor
from ..config import BeamConfig
from enum import Enum


class TransformStrategy(Enum):
    CC = "CC"
    CS = "CS"
    SC = "SC"
    SS = "SS"
    C = "C"
    S = "S"


class Transformer(Processor):

    def __init__(self, *args, n_workers=0, n_chunks=None, name=None, store_path=None, partition=None,
                 chunksize=None, mp_method='joblib', squeeze=True, reduce=True, reduce_dim=0,
                 transform_strategy=None, split_by='keys', store_suffix=None, shuffle=False, **kwargs):
        """

        @param args:
        @param n_workers:
        @param n_chunks:
        @param name:
        @param store_path:
        @param chunksize:
        @param mp_method:
        @param squeeze:
        @param reduce_dim:
        @param transform_strategy: Determines the strategy of cache/store operations during transformation:
            'CC' - the data is cached before the split into multiple chunks and the split to multiprocess,
            the output of each process remains cached and is returned to the main process as a list of cached data.
            'CS' - the data is cached before the split into multiple chunks and the split to multiprocess,
            the output of each process is stored and is returned to the main process as a list of paths.
            This approach suits for enriching the data with additional information, e.g. embeddings
            where the transformed data does not fit into the memory.
            'SC' - the data stored and given to the transformer as a list of paths, the output of each process remains
            cached and is returned to the main process as a list of cached data. This approach suits for the case
            when the input data is too large to fit into the memory but the transformation generate a small output
            that can be cached, e.g. aggregation operations.
            'SS' - the data stored and given to the transformer as a list of paths, the output of each process is stored
            and is returned to the main process as a list of paths. This approach suits for the case when the input data
            is too large to fit into the memory and the transformation generate a large output that cannot be cached,
            e.g. image transformations.
            'C' - the input type is inferred from the BeamData object and the output is cached.
            'S' - the input type is inferred from the BeamData object and the output is stored.
        @param split_by: The split strategy of the data into chunks.
        'keys' - the data is split by the key,
        'index' - the data is split by the index (i.e. dim=0).
        'columns' - the data is split by the columns (i.e. dim=1).
        @param store_suffix: The suffix of the stored file.
        @param shuffle Shuffling the tasks before running them.
        @param kwargs:
        """
        super(Transformer, self).__init__(*args, name=name, **kwargs)

        if (n_chunks is None) and (chunksize is None):
            n_chunks = 1

        self.transformers = None

        if len(args) > 0 and isinstance(args[0], BeamConfig):
            self.hparams = args[0]
        else:
            self.hparams = BeamConfig(chunksizes=chunksize, n_chunks=n_chunks, n_workers=n_workers, squeeze=squeeze,
                                      split_by=split_by, partition=partition, mp_method=mp_method, shuffle=shuffle,
                                      reduce_dim=reduce_dim, transform_strategy=transform_strategy,
                                      reduce=reduce, **kwargs)

        self.chunksize = self.get_hparam('chunksize', preferred=chunksize)
        self.n_chunks = self.get_hparam('n_chunks', preferred=n_chunks)
        self.n_workers = self.get_hparam('n_workers', preferred=n_workers)
        self.squeeze = self.get_hparam('squeeze', preferred=squeeze)
        self.split_by = self.get_hparam('split_by', preferred=split_by)
        self.store_suffix = self.get_hparam('store_suffix', preferred=store_suffix)
        self.transform_strategy = self.get_hparam('transform_strategy', preferred=transform_strategy)
        self.shuffle = self.get_hparam('shuffle', preferred=shuffle)
        self.kwargs = kwargs
        if self.transform_strategy in [TransformStrategy.SC, TransformStrategy.SS] and self.split_by != 'keys':
            logger.warning(f'transformation strategy {self.transform_strategy} supports only split_by=\"keys\", '
                           f'The split_by is set to "keys".')
            self.split_by = 'keys'

        store_path = self.get_hparam('store_path', preferred=store_path)
        if store_path is not None:
            store_path = beam_path(store_path)
        if store_path is not None and name is not None:
            store_path = store_path.joinpath(name)

        self.store_path = store_path
        self.partition = self.get_hparam('partition', preferred=partition)
        self.mp_method = self.get_hparam('mp_method', preferred=mp_method)
        self.reduce_dim = self.get_hparam('reduce_dim', preferred=reduce_dim)
        self.to_reduce = self.get_hparam('reduce', preferred=reduce)
        self._exceptions = None

    def chunks(self, x, chunksize=None, n_chunks=None, squeeze=None, split_by=None, partition=None):

        split_by = split_by or self.split_by
        partition = partition or self.partition

        if (chunksize is None) and (n_chunks is None):
            chunksize = self.chunksize
            n_chunks = self.n_chunks
        if squeeze is None:
            squeeze = self.squeeze

        if isinstance(x, BeamData):
            for k, c in x.divide_chunks(chunksize=chunksize, n_chunks=n_chunks, partition=partition, split_by=split_by):
                yield k, c

        else:

            dim = 0 if split_by == 'index' else 1 if split_by == 'column' else None
            for k, c in recursive_chunks(x, chunksize=chunksize, n_chunks=n_chunks, squeeze=squeeze, dim=dim):
                yield k, c

    def transform_callback(self, x, key=None, is_chunk=False, fit=False, path=None, **kwargs):
        raise NotImplementedError

    def worker(self, x, key=None, is_chunk=False, fit=False, cache=True, store_path=None, **kwargs):

        if isinstance(x, BeamData):
            if not x.is_cached and cache:
                x.cache()

        x = self.transform_callback(x, key=key, is_chunk=is_chunk, fit=fit, **kwargs)

        if store_path is not None:
            store_path = beam_path(store_path)
            if store_path.suffix:
                store_path.write(x)
            else:
                if not isinstance(x, BeamData):
                    x = BeamData(x)
                x.store(path=store_path, )
                x = BeamData.from_path(path=store_path)

        return key, x

    def fit(self, x, **kwargs):
        return x

    @property
    def exceptions(self):
        return self._exceptions

    @exceptions.setter
    def exceptions(self, exceptions):
        self._exceptions = exceptions

    def fit_transform(self, x, **kwargs):
        return self.transform(x, fit=True, **kwargs)
        # self.fit(x, **kwargs)
        # return self.transform(x, **kwargs)

    def reduce(self, x, reduce_dim=None, split_by=None, squeeze=True, **kwargs):

        if isinstance(next(iter_container(x))[1], BeamData):
            x = BeamData.collate(x, split_by=split_by, **kwargs)
        else:

            if reduce_dim is None:
                reduce_dim = self.reduce_dim

            x = collate_chunks(*x, dim=reduce_dim, squeeze=squeeze, **kwargs)

        return x

    def transform(self, x, transform_kwargs=None, parallel_kwargs=None, **kwargs):

        transform_kwargs = transform_kwargs or {}

        split_by = transform_kwargs.pop('split_by', self.split_by)
        partition = transform_kwargs.pop('partition', self.partition)
        mp_method = transform_kwargs.pop('mp_method', self.mp_method)
        shuffle = transform_kwargs.pop('shuffle', self.shuffle)
        n_workers = transform_kwargs.pop('n_workers', self.n_workers)
        store_suffix = transform_kwargs.pop('store_suffix', self.store_suffix)
        transform_strategy = transform_kwargs.pop('transform_strategy', self.transform_strategy)
        reduce = transform_kwargs.pop('reduce', self.to_reduce)
        parallel_kwargs = parallel_kwargs or {}

        reduce_dim = self.reduce_dim

        if transform_strategy in [TransformStrategy.SC, TransformStrategy.SS] and split_by != 'keys':
            logger.warning(f'transformation strategy {transform_strategy} supports only split_by=\"key\", '
                           f'The split_by is set to "key".')
            split_by = 'keys'

        path = transform_kwargs.pop('path', self.store_path)
        store = transform_kwargs.pop('store', (path is not None))

        logger.info(f"Starting transformer process: {self.name}")

        if is_empty(x):
            return x

        chunksize = transform_kwargs.pop('chunksize', self.chunksize)
        n_chunks = transform_kwargs.pop('n_chunks', self.n_chunks)
        squeeze = transform_kwargs.pop('squeeze', self.squeeze)
        if (chunksize is None) and (n_chunks is None):
            chunksize = self.chunksize
            n_chunks = self.n_chunks
        if (chunksize is None) and (n_chunks is None):
            n_chunks = 1
        if squeeze is None:
            squeeze = self.squeeze

        is_chunk = (n_chunks != 1) or (not squeeze) or (split_by == 'keys' and isinstance(x, BeamData) and x.is_stored)

        if ((transform_strategy is None) or (transform_strategy == TransformStrategy.C)) and type(x) == BeamData:
            if x.is_cached:
                transform_strategy = TransformStrategy.CC
            elif x.is_stored:
                transform_strategy = TransformStrategy.SC
            else:
                raise ValueError(f"BeamData is not cached or stored, check your configuration")

        if transform_strategy == TransformStrategy.S and type(x) == BeamData:
            if x.is_cached:
                transform_strategy = TransformStrategy.CS
            elif x.is_stored:
                transform_strategy = TransformStrategy.SS
            else:
                raise ValueError(f"BeamData is not cached or stored, check your configuration")

        if (transform_strategy in [TransformStrategy.CC, TransformStrategy.CS] and
                type(x) == BeamData and not x.is_cached):
            logger.warning(f"Data is not cached but the transformation strategy is {transform_strategy}, "
                           f"caching data for transformer: {self.name} before the split to chunks.")
            x.cache()

        if (transform_strategy in [TransformStrategy.SC, TransformStrategy.SS] and
                type(x) == BeamData and not x.is_stored):
            logger.warning(f"Data is not stored but the transformation strategy is {transform_strategy}, "
                           f"storing data for transformer: {self.name} before the split to chunks.")
            x.store()

        store_chunk = transform_strategy in [TransformStrategy.CS, TransformStrategy.SS]

        if path is None and store_chunk:

            if isinstance(x, BeamData) and x.path is not None:
                path = x.path
                path = path.parent.joinpath(f"{path.name}_transformed_{self.name}")
                logger.info(f"Path is not specified for transformer: {self.name}, "
                            f"the chunk will be stored in a neighboring directory as the original data: {x.path}"
                            f"to: {path}.")
            else:
                logger.warning(f"Path is not specified for transformer: {self.name}, "
                               f"the chunk will not be stored.")
                store_chunk = False
        elif store_chunk:
            logger.info(f"Storing transformed chunks of data in: {path}")

        queue = BeamParallel(n_workers=n_workers, func=None, method=mp_method, name=self.name,
                             progressbar='beam', reduce=False, reduce_dim=reduce_dim, **parallel_kwargs)

        if is_chunk:
            logger.info(f"Splitting data to chunks for transformer: {self.name}")
            for k, c in tqdm(self.chunks(x, chunksize=chunksize, n_chunks=n_chunks,
                                         squeeze=squeeze, split_by=split_by, partition=partition)):

                chunk_path = None
                if store_chunk:

                    if split_by == 'index':
                        part_name = BeamData.index_partition_directory_name
                    elif split_by == 'columns':
                        part_name = BeamData.columns_partition_directory_name
                    else:
                        part_name = ''

                    chunk_path = path.joinpath(f"{BeamData.normalize_key(k)}{part_name}")
                    if store_suffix is not None:
                        chunk_path = f"{chunk_path}{store_suffix}"
                    # chunk_path = chunk_path.as_uri()

                queue.add(BeamTask(self.worker, c, key=k, is_chunk=is_chunk, store_path=chunk_path,
                                   store=store_chunk, name=k, metadata=f"{self.name}", **kwargs))

        else:
            queue.add(BeamTask(self.worker, x, key=None, is_chunk=is_chunk,
                               store=store_chunk, name=self.name, **kwargs))

        logger.info(f"Starting transformer: {self.name} with {n_workers} workers. "
                    f"Number of queued tasks is {len(queue)}.")

        synced_results = queue.run(n_workers=n_workers, method=mp_method, shuffle=shuffle)

        exceptions = []
        for i, (_, v) in enumerate(iter_container(synced_results.exceptions)):
            exceptions.append({'exception': v, 'task': queue.queue[i]})

        if len(exceptions) > 0:
            logger.error(f"Transformer {self.name} had {len(exceptions)} exceptions during operation.")
            logger.info("Failed tasks can be obtained in self.exceptions")
            self.exceptions = exceptions

        results = []
        for _, v in iter_container(synced_results.values):
            if not isinstance(v, Exception):
                results.append(v)

        if is_chunk:
            values = [xi[1] for xi in results]
            keys = [xi[0] for xi in results]
            keys = [ki if type(ki) is tuple else (ki,) for ki in keys]

            if len(exceptions) == 0:
                x = build_container_from_tupled_keys(keys, values)

                logger.info(f"Finished transformer process: {self.name}. Collating results...")

                if reduce:
                    x = self.reduce(x, split_by=split_by, **kwargs)
            else:
                x = {k[0] if type(k) is tuple and len(k) == 1 else k: v for k, v in zip(keys, values)}
                if store:
                    logger.warning("Due to exceptions, the data will not be stored, "
                                   "the data is returned as a dictionary of all the successful tasks.")
                return x

        else:
            if len(exceptions) > 0:
                logger.warning("Exception occurred, the exception object and the task are returned.")
                return results
            logger.info(f"Finished transformer process: {self.name}.")
            x = results[0][1]

        if store:

            logger.info(f"Storing transformed of data in: {path}")
            if not isinstance(x, BeamData):
                x = BeamData(x)
            x.store(path=path)
            x = BeamData.from_path(path=path)

        return x





from ..core.processor import Processor
from ..data import BeamData
from ..path import beam_key
import pandas as pd
import re
from sqlalchemy.engine import create_engine


class BeamSQL(Processor):

    # Build a beam class that provides an abstraction layer to different databases and lets us develop our tools without committing to a database technology.
    #
    # The class will be based on sqlalchemy+pandas but it can be inherited by subclasses that use 3rd party packages such as pyathena.
    #
    # some key features:
    # 1. the interface will be based on url addresses as in the BeamPath class
    # 2. two levels will be supported, db level where each index is a table and table level where each index is a column.
    # 3. minimizing the use of schemas and inferring the schemas from existing pandas dataframes and much as possible
    # 4. adding pandas like api whenever possible, for example, selecting columns with __getitem__, uploading columns and tables with __setitem__, loc, iloc
    # 5. the use of sqlalchemy and direct raw sql queries will be allowed.

    def __init__(self, *args, llm=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = None
        self._engine = None
        self._table = None
        self._database = None
        self._index = None
        self._columns = None
        self._llm = llm

    @property
    def llm(self):
        return self._llm

    @property
    def database(self):
        return self._database

    @property
    def table(self):
        return self._table

    @property
    def index(self):
        return self._index

    @property
    def columns(self):
        return self._columns

    def set_database(self, database):
        self._database = database

    def set_llm(self, llm):
        self._llm = llm

    def set_index(self, index):
        self._index = index

    def set_columns(self, columns):
        self._columns = columns

    def set_table(self, table):
        self._table = table

    def __getitem__(self, item):

        if not isinstance(item, tuple):
            item = (item,)

        if self.table is None:
            axes = ['table', 'index', 'columns']
        else:
            axes = ['index', 'columns']

        for i, ind_i in enumerate(item):
            a = axes.pop(0)
            if a == 'table':
                self.set_table(ind_i)
            elif a == 'index':
                self.set_index(ind_i)
            elif a == 'columns':
                self.set_columns(ind_i)

        return self

    def sql(self, query, **kwargs):
        return pd.read_sql(query, self._connection, **kwargs)

    def get_sample(self, n=1, **kwargs):
        raise NotImplementedError

    def get_schema(self):
        raise NotImplementedError

    def nlp(self, query, **kwargs):

        schema = self.get_schema()

        prompt = f"Task: generate an SQL query that best describes the following text:\n {query}\n\n" \
                 f"++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" \
                 f"Additional instructions:\n\n" \
                 f"1. The queried table name is: {self.database}.{self.table}\n" \
                 f"2. Assume that the schema for the queried table is:\n{schema}\n\n" \
                 f"3. Here are 4 example rows {self.get_sample(n=4)}\n\n" \
                 f"4. In your response use only valid column names that best match the text\n\n" \
                 f"5. Important: your response must contain only the SQL query and nothing else, and it must be valid.\n\n" \
                 f"++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n" \
                 f"Response: \"\"\"\n{{text input here}}\n\"\"\""

        response = self.llm.ask(prompt, **kwargs)

        query = response.choices[0].text
        query = re.sub(r'\"\"\"', '', query)

        return self.sql(query)

    @staticmethod
    def df2table(df, name, metadata=None):

        from sqlalchemy import Table, Column
        from sqlalchemy.schema import MetaData

        if metadata is None:
            metadata = MetaData()

        # Define the SQLAlchemy table object based on the DataFrame
        columns = [column for column in df.columns]
        types = {column: df.dtypes[column].name for column in df.columns}
        table = Table(name, metadata, *(Column(column, types[column]) for column in columns))

        return table

    @property
    def engine(self):
        raise NotImplementedError

    def __enter__(self):
        self._connection = self.engine.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._connection.close()
        self._connection = None


class BeamAthena(BeamSQL):
    def __init__(self, s3_staging_dir, role_session_name=None, region_name=None, access_key=None, secret_key=None,
                 *args, **kwargs):

        self.access_key = beam_key('AWS_ACCESS_KEY_ID', access_key)
        self.secret_key = beam_key('aws_secret_key', secret_key)
        self.s3_staging_dir = s3_staging_dir

        if role_session_name is None:
            role_session_name = "PyAthena-session"
        self.role_session_name = role_session_name

        if region_name is None:
            region_name = "eu-north-1"
        self.region_name = region_name

        state = {'s3_staging_dir': self.s3_staging_dir, 'role_session_name': self.role_session_name,
                      'region_name': self.region_name, 'access_key': self.access_key, 'secret_key': self.secret_key}

        super().__init__(*args, state=state, **kwargs)

    def get_sample(self, n=1, **kwargs):

        from pyathena.pandas.util import as_pandas

        query = f"SELECT * FROM {self.database}.{self.table} LIMIT {n}"

        cursor = self.connection.cursor()
        cursor.execute(query)
        df = as_pandas(cursor)

        return df

    def get_schema(self):

        query = f"DESCRIBE {self.database}.{self.table}"

        cursor = self.connection.cursor()
        cursor.execute(query)
        # Fetch the result
        result = cursor.fetchall()

        return result

    @property
    def engine(self):
        if self._engine is None:
            self._engine = create_engine('athena+pyathena://', creator=lambda: self.connection)
        return self._engine

    @property
    def connection(self):

        if self._connection is None:

            from pyathena import connect

            self._connection = connect(s3_staging_dir=self.s3_staging_dir,
                                       role_session_name=self.role_session_name,
                                       region_name=self.region_name, aws_access_key_id=self.access_key,
                                       aws_secret_access_key=self.secret_key)

        return self._connection

    def sql(self, query):

        from pyathena.pandas.util import as_pandas

        cursor = self.connection.cursor()
        cursor.execute(query)
        df = as_pandas(cursor)
        bd = BeamData(df)

        return bd


# from sqlalchemy import create_engine, func, MetaData, Table, and_
# import pandas as pd
#
#
# class BeamSQL:
#     def __init__(self, uri):
#         self.engine = create_engine(uri)
#         self.metadata = MetaData(bind=self.engine)
#         self.current_table = None
#         self.groupby_columns = []
#         self.filter_conditions = []
#
#     def table(self, table_name):
#         """Select the current working table."""
#         self.current_table = Table(table_name, self.metadata, autoload=True)
#         return self
#
#     def groupby(self, *column_names):
#         self.groupby_columns = column_names
#         return self
#
#     def count(self):
#         if not self.current_table or not self.groupby_columns:
#             raise Exception("Table not selected or columns for grouping not provided")
#
#         columns_to_select = [getattr(self.current_table.c, col) for col in self.groupby_columns]
#         s = select(columns_to_select + [func.count()]) \
#             .where(and_(*self.filter_conditions)) \
#             .group_by(*columns_to_select)
#
#         result = self.engine.execute(s)
#         return pd.DataFrame(result.fetchall(), columns=self.groupby_columns + ['count'])
#
#     def filter(self, column_name, operator, value):
#         column = getattr(self.current_table.c, column_name)
#         op_map = {
#             "==": column.__eq__,
#             ">": column.__gt__,
#             "<": column.__lt__,
#             ">=": column.__ge__,
#             "<=": column.__le__,
#         }
#
#         if operator not in op_map:
#             raise Exception(f"Operator {operator} not supported")
#
#         condition = op_map[operator](value)
#         self.filter_conditions.append(condition)
#         return self
#
#     def query(self, raw_sql):
#         result = self.engine.execute(raw_sql)
#         columns = result.keys()
#         return pd.DataFrame(result.fetchall(), columns=columns)
#
#
# # Sample usage:
# bs = BeamSQL("sqlite:////path/to/sqlite3.db")
#
# # Equivalent to df.groupby(['name', 'age']).count()
# result = bs.table('your_table_name').groupby('name', 'age').count()
# print(result)
#
# # Equivalent to df[df['age'] > 25].groupby(['name']).count()
# result = bs.table('your_table_name').filter('age', '>', 25).groupby('name').count()
# print(result)
#
# # Raw SQL
# result = bs.query("SELECT name, COUNT(*) FROM your_table_name GROUP BY name")
# print(result)



from ..config import UniversalConfig, BeamParam


class FTLLMConfig(UniversalConfig):

    defaults = dict(project_name='fine_tune_llm', training_framework='accelerate', batch_size=2,
                    model_dtype='bfloat16', epoch_length=100,
                    scale_epoch_by_batch_size=False, lr_dense=1e-5, lr_sparse=1e-4, reduction='mean_batch')
    parameters = [BeamParam('model', str, None, 'Model to use for fine-tuning'),
                  BeamParam('prompt_key', str, 'prompt', 'Key to use for the prompt'),
                  BeamParam('completion_key', str, None, 'Key to use for the completion'),
                  BeamParam('lora_alpha', float, 16, 'Lora alpha parameter', tags=['tune']),
                  BeamParam('lora_dropout', float, 0.05, 'Lora dropout', tags=['tune']),
                  BeamParam('lora_r', int, 16, 'Lora r parameter', tags=['tune']),
                  BeamParam('lora_fan_in_fan_out', bool, False, 'Set this to True if the layer to replace stores '
                                                                'weight like (fan_in, fan_out)'),
                  BeamParam('lora_bias', str, 'none', 'Bias type for Lora. Can be ‘none’, '
                                                      '‘all’ or ‘lora_only’. If ‘all’ or ‘lora_only’', tags=['tune']),
                  BeamParam('load_in_8bit', bool, False, 'Load the model in 8bit mode'),
                  BeamParam('modules_to_save', list, None, 'List of modules apart from LoRA layers to be set '
                                                           'as trainable and saved in the final checkpoint'),
                  BeamParam('layers_to_transform', list, None, 'The layer indexes to transform, if this argument '
                                                               'is specified, it will apply the LoRA transformations '
                                                               'on the layer indexes that are specified in this list.'),
                  BeamParam('target_modules', list, None, 'The names of the modules to apply Lora to'),
                  BeamParam('hf_cache_dir', str, None, 'Directory for Huggingface to cache to and load from'),
                  BeamParam('hf_data_dir', str, None, 'Directory for the dataset to load from'),
                  BeamParam('return_overflowing_tokens', bool, False, 'Whether or not to split overflowing tokens into '
                                                                      'their own batch'),
                  BeamParam('context_length', int, 128, 'The maximal context length to train the model with',
                            tags=['tune']),

                  BeamParam('dataset', str, None, 'The dataset which is used for fine-tuning'),
                  ]
